@@ -8,30 +8,30 @@ import (
 	"sync"
 	"time"
 
+	"github.com/aws/aws-sdk-go/aws"
 	"github.com/aws/aws-sdk-go/aws/awserr"
 	"github.com/aws/aws-sdk-go/aws/awsutil"
+	"github.com/aws/aws-sdk-go/aws/client"
+	"github.com/aws/aws-sdk-go/aws/request"
 	"github.com/aws/aws-sdk-go/service/s3"
+	"github.com/aws/aws-sdk-go/service/s3/s3iface"
 )
 
-// The maximum allowed number of parts in a multi-part upload on Amazon S3.
-var MaxUploadParts = 10000
+// MaxUploadParts is the maximum allowed number of parts in a multi-part upload
+// on Amazon S3.
+const MaxUploadParts = 10000
 
-// The minimum allowed part size when uploading a part to Amazon S3.
-var MinUploadPartSize int64 = 1024 * 1024 * 5
+// MinUploadPartSize is the minimum allowed part size when uploading a part to
+// Amazon S3.
+const MinUploadPartSize int64 = 1024 * 1024 * 5
 
-// The default part size to buffer chunks of a payload into.
-var DefaultUploadPartSize = MinUploadPartSize
+// DefaultUploadPartSize is the default part size to buffer chunks of a
+// payload into.
+const DefaultUploadPartSize = MinUploadPartSize
 
-// The default number of goroutines to spin up when using Upload().
-var DefaultUploadConcurrency = 5
-
-// The default set of options used when opts is nil in Upload().
-var DefaultUploadOptions = &UploadOptions{
-	PartSize:          DefaultUploadPartSize,
-	Concurrency:       DefaultUploadConcurrency,
-	LeavePartsOnError: false,
-	S3:                nil,
-}
+// DefaultUploadConcurrency is the default number of goroutines to spin up when
+// using Upload().
+const DefaultUploadConcurrency = 5
 
 // A MultiUploadFailure wraps a failed S3 multipart upload. An error returned
 // will satisfy this interface when a multi part upload failed to upload all
@@ -43,7 +43,7 @@ var DefaultUploadOptions = &UploadOptions{
 //     u := s3manager.NewUploader(opts)
 //     output, err := u.upload(input)
 //     if err != nil {
-//         if multierr, ok := err.(MultiUploadFailure); ok {
+//         if multierr, ok := err.(s3manager.MultiUploadFailure); ok {
 //             // Process error and its associated uploadID
 //             fmt.Println("Error:", multierr.Code(), multierr.Message(), multierr.UploadID())
 //         } else {
@@ -166,7 +166,7 @@ type UploadInput struct {
 	// requests for an object protected by AWS KMS will fail if not made via SSL
 	// or using SigV4. Documentation on configuring any of the officially supported
 	// AWS SDKs and CLI can be found at http://docs.aws.amazon.com/AmazonS3/latest/dev/UsingAWSSDK.html#specify-signature-version
-	SSEKMSKeyID *string `location:"header" locationName:"x-amz-server-side-encryption-aws-kms-key-id" type:"string"`
+	SSEKMSKeyId *string `location:"header" locationName:"x-amz-server-side-encryption-aws-kms-key-id" type:"string"`
 
 	// The Server-side encryption algorithm used when storing this object in S3
 	// (e.g., AES256, aws:kms).
@@ -189,20 +189,27 @@ type UploadOutput struct {
 	// The URL where the object was uploaded to.
 	Location string
 
+	// The version of the object that was uploaded. Will only be populated if
+	// the S3 Bucket is versioned. If the bucket is not versioned this field
+	// will not be set.
+	VersionID *string
+
 	// The ID for a multipart upload to S3. In the case of an error the error
 	// can be cast to the MultiUploadFailure interface to extract the upload ID.
 	UploadID string
 }
 
-// UploadOptions keeps tracks of extra options to pass to an Upload() call.
-type UploadOptions struct {
+// The Uploader structure that calls Upload(). It is safe to call Upload()
+// on this structure for multiple objects and across concurrent goroutines.
+// Mutating the Uploader's properties is not safe to be done concurrently.
+type Uploader struct {
 	// The buffer size (in bytes) to use when buffering data into chunks and
 	// sending them as parts to S3. The minimum allowed part size is 5MB, and
 	// if this value is set to zero, the DefaultPartSize value will be used.
 	PartSize int64
 
 	// The number of goroutines to spin up in parallel when sending parts.
-	// If this is set to zero, the DefaultConcurrency value will be used.
+	// If this is set to zero, the DefaultUploadConcurrency value will be used.
 	Concurrency int
 
 	// Setting this value to true will cause the SDK to avoid calling
@@ -213,43 +220,124 @@ type UploadOptions struct {
 	// space usage on S3 and will add additional costs if not cleaned up.
 	LeavePartsOnError bool
 
-	// The client to use when uploading to S3. Leave this as nil to use the
-	// default S3 client.
-	S3 *s3.S3
+	// MaxUploadParts is the max number of parts which will be uploaded to S3.
+	// Will be used to calculate the partsize of the object to be uploaded.
+	// E.g: 5GB file, with MaxUploadParts set to 100, will upload the file
+	// as 100, 50MB parts.
+	// With a limited of s3.MaxUploadParts (10,000 parts).
+	MaxUploadParts int
+
+	// The client to use when uploading to S3.
+	S3 s3iface.S3API
 }
 
-// NewUploader creates a new Uploader object to upload data to S3. Pass in
-// an optional opts structure to customize the uploader behavior.
-func NewUploader(opts *UploadOptions) *Uploader {
-	if opts == nil {
-		opts = DefaultUploadOptions
+// NewUploader creates a new Uploader instance to upload objects to S3. Pass In
+// additional functional options to customize the uploader's behavior. Requires a
+// client.ConfigProvider in order to create a S3 service client. The session.Session
+// satisfies the client.ConfigProvider interface.
+//
+// Example:
+//     // The session the S3 Uploader will use
+//     sess, err := session.NewSession()
+//
+//     // Create an uploader with the session and default options
+//     uploader := s3manager.NewUploader(sess)
+//
+//     // Create an uploader with the session and custom options
+//     uploader := s3manager.NewUploader(session, func(u *s3manager.Uploader) {
+//          u.PartSize = 64 * 1024 * 1024 // 64MB per part
+//     })
+func NewUploader(c client.ConfigProvider, options ...func(*Uploader)) *Uploader {
+	u := &Uploader{
+		S3:                s3.New(c),
+		PartSize:          DefaultUploadPartSize,
+		Concurrency:       DefaultUploadConcurrency,
+		LeavePartsOnError: false,
+		MaxUploadParts:    MaxUploadParts,
 	}
-	return &Uploader{opts: opts}
+
+	for _, option := range options {
+		option(u)
+	}
+
+	return u
 }
 
-// The Uploader structure that calls Upload(). It is safe to call Upload()
-// on this structure for multiple objects and across concurrent goroutines.
-type Uploader struct {
-	opts *UploadOptions
+// NewUploaderWithClient creates a new Uploader instance to upload objects to S3. Pass in
+// additional functional options to customize the uploader's behavior. Requires
+// a S3 service client to make S3 API calls.
+//
+// Example:
+//     // The session the S3 Uploader will use
+//     sess, err := session.NewSession()
+//
+//     // S3 service client the Upload manager will use.
+//     s3Svc := s3.New(sess)
+//
+//     // Create an uploader with S3 client and default options
+//     uploader := s3manager.NewUploaderWithClient(s3Svc)
+//
+//     // Create an uploader with S3 client and custom options
+//     uploader := s3manager.NewUploaderWithClient(s3Svc, func(u *s3manager.Uploader) {
+//          u.PartSize = 64 * 1024 * 1024 // 64MB per part
+//     })
+func NewUploaderWithClient(svc s3iface.S3API, options ...func(*Uploader)) *Uploader {
+	u := &Uploader{
+		S3:                svc,
+		PartSize:          DefaultUploadPartSize,
+		Concurrency:       DefaultUploadConcurrency,
+		LeavePartsOnError: false,
+		MaxUploadParts:    MaxUploadParts,
+	}
+
+	for _, option := range options {
+		option(u)
+	}
+
+	return u
 }
 
 // Upload uploads an object to S3, intelligently buffering large files into
 // smaller chunks and sending them in parallel across multiple goroutines. You
-// can configure the buffer size and concurrency through the opts parameter.
+// can configure the buffer size and concurrency through the Uploader's parameters.
 //
-// If opts is set to nil, DefaultUploadOptions will be used.
+// Additional functional options can be provided to configure the individual
+// upload. These options are copies of the Uploader instance Upload is called from.
+// Modifying the options will not impact the original Uploader instance.
 //
-// It is safe to call this method for multiple objects and across concurrent
-// goroutines.
-func (u *Uploader) Upload(input *UploadInput) (*UploadOutput, error) {
-	i := uploader{in: input, opts: *u.opts}
+// It is safe to call this method concurrently across goroutines.
+//
+// Example:
+//     // Upload input parameters
+//     upParams := &s3manager.UploadInput{
+//         Bucket: &bucketName,
+//         Key:    &keyName,
+//         Body:   file,
+//     }
+//
+//     // Perform an upload.
+//     result, err := uploader.Upload(upParams)
+//
+//     // Perform upload with options different than the those in the Uploader.
+//     result, err := uploader.Upload(upParams, func(u *s3manager.Uploader) {
+//          u.PartSize = 10 * 1024 * 1024 // 10MB part size
+//          u.LeavePartsOnError = true    // Don't delete the parts if the upload fails.
+//     })
+func (u Uploader) Upload(input *UploadInput, options ...func(*Uploader)) (*UploadOutput, error) {
+	i := uploader{in: input, ctx: u}
+
+	for _, option := range options {
+		option(&i.ctx)
+	}
+
 	return i.upload()
 }
 
 // internal structure to manage an upload to S3.
 type uploader struct {
-	in   *UploadInput
-	opts UploadOptions
+	ctx Uploader
+
+	in *UploadInput
 
 	readerPos int64 // current reader position
 	totalSize int64 // set to -1 if the size is not known
@@ -260,33 +348,30 @@ type uploader struct {
 func (u *uploader) upload() (*UploadOutput, error) {
 	u.init()
 
-	if u.opts.PartSize < MinUploadPartSize {
+	if u.ctx.PartSize < MinUploadPartSize {
 		msg := fmt.Sprintf("part size must be at least %d bytes", MinUploadPartSize)
 		return nil, awserr.New("ConfigError", msg, nil)
 	}
 
 	// Do one read to determine if we have more than one part
-	buf, err := u.nextReader()
-	if err == io.EOF || err == io.ErrUnexpectedEOF { // single part
-		return u.singlePart(buf)
+	reader, _, err := u.nextReader()
+	if err == io.EOF { // single part
+		return u.singlePart(reader)
 	} else if err != nil {
 		return nil, awserr.New("ReadRequestBody", "read upload data failed", err)
 	}
 
 	mu := multiuploader{uploader: u}
-	return mu.upload(buf)
+	return mu.upload(reader)
 }
 
 // init will initialize all default options.
 func (u *uploader) init() {
-	if u.opts.S3 == nil {
-		u.opts.S3 = s3.New(nil)
+	if u.ctx.Concurrency == 0 {
+		u.ctx.Concurrency = DefaultUploadConcurrency
 	}
-	if u.opts.Concurrency == 0 {
-		u.opts.Concurrency = DefaultUploadConcurrency
-	}
-	if u.opts.PartSize == 0 {
-		u.opts.PartSize = DefaultUploadPartSize
+	if u.ctx.PartSize == 0 {
+		u.ctx.PartSize = DefaultUploadPartSize
 	}
 
 	// Try to get the total size for some optimizations
@@ -309,9 +394,12 @@ func (u *uploader) initSize() {
 		}
 		u.totalSize = n
 
-		// try to adjust partSize if it is too small
-		if u.totalSize/u.opts.PartSize >= int64(MaxUploadParts) {
-			u.opts.PartSize = u.totalSize / int64(MaxUploadParts)
+		// Try to adjust partSize if it is too small and account for
+		// integer division truncation.
+		if u.totalSize/u.ctx.PartSize >= int64(u.ctx.MaxUploadParts) {
+			// Add one to the part size to account for remainders
+			// during the size calculation. e.g odd number of bytes.
+			u.ctx.PartSize = (u.totalSize / int64(u.ctx.MaxUploadParts)) + 1
 		}
 	}
 }
@@ -320,38 +408,49 @@ func (u *uploader) initSize() {
 // This operation increases the shared u.readerPos counter, but note that it
 // does not need to be wrapped in a mutex because nextReader is only called
 // from the main thread.
-func (u *uploader) nextReader() (io.ReadSeeker, error) {
+func (u *uploader) nextReader() (io.ReadSeeker, int, error) {
+	type readerAtSeeker interface {
+		io.ReaderAt
+		io.ReadSeeker
+	}
 	switch r := u.in.Body.(type) {
-	case io.ReaderAt:
+	case readerAtSeeker:
 		var err error
 
-		n := u.opts.PartSize
+		n := u.ctx.PartSize
 		if u.totalSize >= 0 {
 			bytesLeft := u.totalSize - u.readerPos
 
-			if bytesLeft == 0 {
+			if bytesLeft <= u.ctx.PartSize {
 				err = io.EOF
 				n = bytesLeft
-			} else if bytesLeft <= u.opts.PartSize {
-				err = io.ErrUnexpectedEOF
-				n = bytesLeft
 			}
 		}
 
-		buf := io.NewSectionReader(r, u.readerPos, n)
+		reader := io.NewSectionReader(r, u.readerPos, n)
 		u.readerPos += n
 
-		return buf, err
+		return reader, int(n), err
 
 	default:
-		packet := make([]byte, u.opts.PartSize)
-		n, err := io.ReadFull(u.in.Body, packet)
+		part := make([]byte, u.ctx.PartSize)
+		n, err := readFillBuf(r, part)
 		u.readerPos += int64(n)
 
-		return bytes.NewReader(packet[0:n]), err
+		return bytes.NewReader(part[0:n]), n, err
 	}
 }
 
+func readFillBuf(r io.Reader, b []byte) (offset int, err error) {
+	for offset < len(b) && err == nil {
+		var n int
+		n, err = r.Read(b[offset:])
+		offset += n
+	}
+
+	return offset, err
+}
+
 // singlePart contains upload logic for uploading a single chunk via
 // a regular PutObject request. Multipart requests require at least two
 // parts, or at least 5MB of data.
@@ -360,13 +459,17 @@ func (u *uploader) singlePart(buf io.ReadSeeker) (*UploadOutput, error) {
 	awsutil.Copy(params, u.in)
 	params.Body = buf
 
-	req, _ := u.opts.S3.PutObjectRequest(params)
+	req, out := u.ctx.S3.PutObjectRequest(params)
+	req.Handlers.Build.PushBack(request.MakeAddToUserAgentFreeFormHandler("S3Manager"))
 	if err := req.Send(); err != nil {
 		return nil, err
 	}
 
 	url := req.HTTPRequest.URL.String()
-	return &UploadOutput{Location: url}, nil
+	return &UploadOutput{
+		Location:  url,
+		VersionID: out.VersionId,
+	}, nil
 }
 
 // internal structure to manage a specific multipart upload to S3.
@@ -400,15 +503,16 @@ func (u *multiuploader) upload(firstBuf io.ReadSeeker) (*UploadOutput, error) {
 	awsutil.Copy(params, u.in)
 
 	// Create the multipart
-	resp, err := u.opts.S3.CreateMultipartUpload(params)
-	if err != nil {
+	req, resp := u.ctx.S3.CreateMultipartUploadRequest(params)
+	req.Handlers.Build.PushBack(request.MakeAddToUserAgentFreeFormHandler("S3Manager"))
+	if err := req.Send(); err != nil {
 		return nil, err
 	}
 	u.uploadID = *resp.UploadId
 
 	// Create the workers
-	ch := make(chan chunk, u.opts.Concurrency)
-	for i := 0; i < u.opts.Concurrency; i++ {
+	ch := make(chan chunk, u.ctx.Concurrency)
+	for i := 0; i < u.ctx.Concurrency; i++ {
 		u.wg.Add(1)
 		go u.readChunk(ch)
 	}
@@ -418,31 +522,43 @@ func (u *multiuploader) upload(firstBuf io.ReadSeeker) (*UploadOutput, error) {
 	ch <- chunk{buf: firstBuf, num: num}
 
 	// Read and queue the rest of the parts
-	for u.geterr() == nil {
+	var err error
+	for u.geterr() == nil && err == nil {
+		num++
 		// This upload exceeded maximum number of supported parts, error now.
-		if num > int64(MaxUploadParts) {
-			msg := fmt.Sprintf("exceeded total allowed parts (%d). "+
-				"Adjust PartSize to fit in this limit", MaxUploadParts)
+		if num > int64(u.ctx.MaxUploadParts) || num > int64(MaxUploadParts) {
+			var msg string
+			if num > int64(u.ctx.MaxUploadParts) {
+				msg = fmt.Sprintf("exceeded total allowed configured MaxUploadParts (%d). Adjust PartSize to fit in this limit",
+					u.ctx.MaxUploadParts)
+			} else {
+				msg = fmt.Sprintf("exceeded total allowed S3 limit MaxUploadParts (%d). Adjust PartSize to fit in this limit",
+					MaxUploadParts)
+			}
 			u.seterr(awserr.New("TotalPartsExceeded", msg, nil))
 			break
 		}
 
-		num++
+		var reader io.ReadSeeker
+		var nextChunkLen int
+		reader, nextChunkLen, err = u.nextReader()
 
-		buf, err := u.nextReader()
-		if err == io.EOF {
-			break
-		}
-
-		ch <- chunk{buf: buf, num: num}
-
-		if err != nil && err != io.ErrUnexpectedEOF {
+		if err != nil && err != io.EOF {
 			u.seterr(awserr.New(
 				"ReadRequestBody",
 				"read multipart upload data failed",
 				err))
 			break
 		}
+
+		if nextChunkLen == 0 {
+			// No need to upload empty part, if file was empty to start
+			// with empty single part would of been created and never
+			// started multipart upload.
+			break
+		}
+
+		ch <- chunk{buf: reader, num: num}
 	}
 
 	// Close the channel, wait for workers, and complete upload
@@ -460,8 +576,9 @@ func (u *multiuploader) upload(firstBuf io.ReadSeeker) (*UploadOutput, error) {
 		}
 	}
 	return &UploadOutput{
-		Location: *complete.Location,
-		UploadID: u.uploadID,
+		Location:  aws.StringValue(complete.Location),
+		VersionID: complete.VersionId,
+		UploadID:  u.uploadID,
 	}, nil
 }
 
@@ -487,15 +604,15 @@ func (u *multiuploader) readChunk(ch chan chunk) {
 // send performs an UploadPart request and keeps track of the completed
 // part information.
 func (u *multiuploader) send(c chunk) error {
-	resp, err := u.opts.S3.UploadPart(&s3.UploadPartInput{
+	req, resp := u.ctx.S3.UploadPartRequest(&s3.UploadPartInput{
 		Bucket:     u.in.Bucket,
 		Key:        u.in.Key,
 		Body:       c.buf,
 		UploadId:   &u.uploadID,
 		PartNumber: &c.num,
 	})
-
-	if err != nil {
+	req.Handlers.Build.PushBack(request.MakeAddToUserAgentFreeFormHandler("S3Manager"))
+	if err := req.Send(); err != nil {
 		return err
 	}
 
@@ -527,15 +644,17 @@ func (u *multiuploader) seterr(e error) {
 
 // fail will abort the multipart unless LeavePartsOnError is set to true.
 func (u *multiuploader) fail() {
-	if u.opts.LeavePartsOnError {
+	if u.ctx.LeavePartsOnError {
 		return
 	}
 
-	u.opts.S3.AbortMultipartUpload(&s3.AbortMultipartUploadInput{
+	req, _ := u.ctx.S3.AbortMultipartUploadRequest(&s3.AbortMultipartUploadInput{
 		Bucket:   u.in.Bucket,
 		Key:      u.in.Key,
 		UploadId: &u.uploadID,
 	})
+	req.Handlers.Build.PushBack(request.MakeAddToUserAgentFreeFormHandler("S3Manager"))
+	req.Send()
 }
 
 // complete successfully completes a multipart upload and returns the response.
@@ -548,13 +667,14 @@ func (u *multiuploader) complete() *s3.CompleteMultipartUploadOutput {
 	// Parts must be sorted in PartNumber order.
 	sort.Sort(u.parts)
 
-	resp, err := u.opts.S3.CompleteMultipartUpload(&s3.CompleteMultipartUploadInput{
+	req, resp := u.ctx.S3.CompleteMultipartUploadRequest(&s3.CompleteMultipartUploadInput{
 		Bucket:          u.in.Bucket,
 		Key:             u.in.Key,
 		UploadId:        &u.uploadID,
 		MultipartUpload: &s3.CompletedMultipartUpload{Parts: u.parts},
 	})
-	if err != nil {
+	req.Handlers.Build.PushBack(request.MakeAddToUserAgentFreeFormHandler("S3Manager"))
+	if err := req.Send(); err != nil {
 		u.seterr(err)
 		u.fail()
 	}
