@@ -13,7 +13,7 @@ var errSSERequiresSSL = awserr.New("ConfigError", "cannot send SSE keys over HTT
 
 func validateSSERequiresSSL(r *request.Request) {
 	if r.HTTPRequest.URL.Scheme != "https" {
-		p := awsutil.ValuesAtPath(r.Params, "SSECustomerKey||CopySourceSSECustomerKey")
+		p, _ := awsutil.ValuesAtPath(r.Params, "SSECustomerKey||CopySourceSSECustomerKey")
 		if len(p) > 0 {
 			r.Error = errSSERequiresSSL
 		}
